__title__ = "mathy"
__version__ = "0.6.5"
__summary__ = "Mathy - RL environments for solving math problems step-by-step"
__uri__ = "https://mathy.ai"
__author__ = "Justin DuJardin"
__email__ = "justin@dujardinconsulting.com"
__license__ = "All rights reserved"
